require 'pry'
require 'net_http_ssl_fix'

require 'chunky_png'
require 'json'
require 'net/http'
require 'net/http/post/multipart'
require 'uri'

module WatchDoge
  module Notification
    class GitlabRepo < WatchDoge::Notification::Base
      def initialize opt
        super

        @host = opt[:host] || ENV['CI_API_V4_URL']
        @project_id = opt[:project_id] || ENV['CI_PROJECT_ID']
        @source_branch = opt[:source_branch] || ENV['CI_COMMIT_REF_NAME']
        @commit_short_sha = opt[:commit_short_sha] || ENV['CI_COMMIT_SHORT_SHA']

        @private_token = opt[:private_token]
      end

      def flush
        return if @message_queue.empty?

        meesages_to_flush = @message_queue
        @message_queue = []

        target_id = get_latest_request_iid || get_commit_id

        raise RuntimeError, "Can't find any target to pushing message" if target_id.nil?

        thread = disscussion_thread do |thread_context|
          meesages_to_flush.each do |message|
            case message
            when String
              post_discussion target_id, message: message
            when ChunkyPNG::Image
              upload_link = upload_image message
              link = md_img_link upload_link

              post_discussion target_id, message: link
            when WatchDoge::PixelTest
              filename = message.option[:filename]

              before = md_img_link upload_image(message.before)
              after = md_img_link upload_image(message.after)
              diff = md_img_link upload_image(message.diff)

              thread_context << markdown_table(before, after, diff, filename)
            end
          end

          thread_context
        end

        post_discussion target_id, message: thread
      end

      def post_discussion target_iid, message:
        uri =
          case @discussion_target
          when :merge_request
            project_uri("/merge_requests/#{target_iid}/discussions")
          when :commit
            project_uri("/repository/commits/#{target_iid}/discussions")
          end

        Net::HTTP.post uri,
                       { body: message }.to_json,
                       "Content-Type": "application/json",
                       "PRIVATE-TOKEN": @private_token
      end

      def upload_image image
        uri = project_uri "/uploads"

        req = Net::HTTP::Post::Multipart.new uri.path, {
          file: UploadIO.new(StringIO.new(image.to_blob), 'image/png', 'image.png')
        }

        req.add_field("PRIVATE-TOKEN", @private_token)

        res = Net::HTTP.start(uri.host, uri.port, use_ssl: (uri.scheme == "https")) do |http|
          res = http.request(req).body
        end

        JSON.parse(res)['url']
      end

      def get_latest_request_iid
        @discussion_target = :merge_request

        uri = project_uri "/merge_requests?source_branch=#{@source_branch}&view=simple&state=opened"

        req = Net::HTTP::Get.new uri
        req.add_field("PRIVATE-TOKEN", @private_token)

        res = Net::HTTP.start(uri.host, uri.port, use_ssl: (uri.scheme == "https")) do |http|
          http.request(req).body
        end

        res = JSON.parse res

        return nil if res.empty?

        res[0]['iid']
      end

      def get_commit_id
        @discussion_target = :commit

        uri = project_uri "/repository/commits/#{@commit_short_sha}"

        req = Net::HTTP::Get.new uri
        req.add_field("PRIVATE-TOKEN", @private_token)

        res = Net::HTTP.start(uri.host, uri.port, use_ssl: (uri.scheme == "https")) do |http|
          http.request(req).body
        end

        res = JSON.parse res

        res['id']
      end

      private

      def md_img_link link
        "![image](#{ENV['CI_PROJECT_URL']}#{link})"
      end

      # disscussion_thread do |context|
      def disscussion_thread
        <<~MARKDOWN
          #### Detail: #{ENV['CI_PIPELINE_URL']}
          #{yield("")}
        MARKDOWN
      end

      def markdown_table before, after, diff, filename
        <<~MARKDOWN
          <h5>#{filename}</h5>
          <table>
            <tr>
              <th>Reference</th>
              <th>Test</th>
              <th>Diff</th>
            </tr>
            <tr>
              <td>#{before}</td>
              <td>#{after}</td>
              <td>#{diff}</td>
            </tr>
          </table>
        MARKDOWN
      end

      def project_uri path
        URI(@host + "/projects/#{@project_id}#{path}")
      end
    end
  end
end
